// $Id: nitrogenbutton.cpp,v 1.35 2009/11/04 20:00:03 hpereira Exp $

//////////////////////////////////////////////////////////////////////////////
// nitrogenbutton.cpp
// -------------------
//
// Copyright (c) 2009 Hugo Pereira Da Costa <hugo.pereira@free.fr>
// Copyright (c) 2006, 2007 Riccardo Iaconelli <riccardo@kde.org>
// Copyright (c) 2006, 2007 Casper Boemann <cbr@boemann.dk>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//////////////////////////////////////////////////////////////////////////////

#include "nitrogenbutton.h"
#include "nitrogenbutton.moc"
#include "nitrogenclient.h"
#include "nitrogen.h"

#include <cmath>

#include <QtGui/QPainter>
#include <QtGui/QPen>

#include <kcommondecoration.h>
#include <KGlobal>
#include <KColorUtils>
#include <KColorScheme>
#include <KDebug>

namespace Nitrogen
{
  //_______________________________________________
  NitrogenButton::NitrogenButton(NitrogenClient &parent,
    const QString& tip, ButtonType type):
    KCommonDecorationButton((::ButtonType)type, &parent),
    client_(parent),
    helper_( parent.helper() ),
    type_(type),
    colorCacheInvalid_(true),
    timeLine_( 200, this )
  {
    setAutoFillBackground(false);
    setAttribute(Qt::WA_OpaquePaintEvent, false);
    setAttribute(Qt::WA_NoSystemBackground);

    {
      unsigned int size( client_.configuration().buttonSize() );
      setFixedSize( size, size );
    }

    setCursor(Qt::ArrowCursor);
    setToolTip(tip);

    timeLine_.setFrameRange( 0, 1000 );
    timeLine_.setCurveShape( QTimeLine::EaseInOutCurve );
    connect( &timeLine_, SIGNAL( frameChanged( int ) ), SLOT( update() ) );
    connect( &timeLine_, SIGNAL( finished() ), SLOT( update() ) );

  }

  //_______________________________________________
  NitrogenButton::~NitrogenButton()
  {}

  //declare function from Nitrogenclient.cpp
  QColor reduceContrast(const QColor &c0, const QColor &c1, double t);

  //_______________________________________________
  QColor NitrogenButton::buttonDetailColor(const QPalette &palette)
  {
    if( client_.timeLineIsRunning() ) return KColorUtils::mix(
      buttonDetailColor( palette, false ),
      buttonDetailColor( palette, true ),
      client_.opacity() );
    else return buttonDetailColor( palette, client_.isActive() );
  }

  //_______________________________________________
  QColor NitrogenButton::buttonDetailColor(const QPalette &palette, bool active)
  {

    if( active ) return palette.color(QPalette::Active, QPalette::WindowText);
    else {

      if (colorCacheInvalid_)
      {
        QColor ab = palette.color(QPalette::Active, QPalette::Window);
        QColor af = palette.color(QPalette::Active, QPalette::WindowText);
        QColor nb = palette.color(QPalette::Inactive, QPalette::Window);
        QColor nf = palette.color(QPalette::Inactive, QPalette::WindowText);

        colorCacheInvalid_ = false;
        cachedButtonDetailColor_ = reduceContrast(nb, nf, qMax(qreal(2.5), KColorUtils::contrastRatio(ab, KColorUtils::mix(ab, af, 0.4))));
      }

      return cachedButtonDetailColor_;
    }
  }

  //___________________________________________________
  QSize NitrogenButton::sizeHint() const
  {
    unsigned int size( client_.configuration().buttonSize() );
    return QSize( size, size );
  }

  //___________________________________________________
  void NitrogenButton::enterEvent(QEvent *e)
  {
    KCommonDecorationButton::enterEvent(e);
    if (status_ != Nitrogen::Pressed) status_ = Nitrogen::Hovered;

    timeLine_.setDirection( QTimeLine::Forward );
    if( !timeLineIsRunning() ) timeLine_.start();

    update();
  }

  //___________________________________________________
  void NitrogenButton::leaveEvent(QEvent *e)
  {
    KCommonDecorationButton::leaveEvent(e);

    if( status_ == Nitrogen::Hovered )
    {
      status_ = Nitrogen::Normal;
      timeLine_.setDirection( QTimeLine::Backward );
      if( !timeLineIsRunning() ) timeLine_.start();
    } else {
      status_ = Nitrogen::Normal;
      update();
    }

  }

  //___________________________________________________
  void NitrogenButton::mousePressEvent(QMouseEvent *e)
  {

    status_ = Nitrogen::Pressed;
    update();

    KCommonDecorationButton::mousePressEvent(e);
  }

  //___________________________________________________
  void NitrogenButton::mouseReleaseEvent(QMouseEvent *e)
  {
    status_ = ( rect().contains( e->pos() ) ) ? Nitrogen::Hovered:Nitrogen::Normal;
    update();

    KCommonDecorationButton::mouseReleaseEvent(e);
  }

  //___________________________________________________
  void NitrogenButton::paintEvent(QPaintEvent *event)
  {

    // do nothing if titlebar is hidden
    if( client_.configuration().hideTitleBar() ) return;

    QPainter painter(this);
    painter.setClipRect(this->rect().intersected( event->rect() ) );
    painter.setRenderHints(QPainter::Antialiasing);

    QPalette palette = ( client_.configuration().drawTitleOutline() || client_.configuration().overwriteColors() ) ? NitrogenButton::palette() : client_.backgroundPalette( this, NitrogenButton::palette() );

    if( client_.isActive() ) palette.setCurrentColorGroup(QPalette::Active);
    else palette.setCurrentColorGroup(QPalette::Inactive);

    // window background
    #if !KDE_IS_VERSION(4,3,0)
    client_.renderWindowBackground( &painter, rect(), this, palette );

    if( client_.isActive() && client_.configuration().drawTitleOutline() && !client_.isMaximized() )
    { client_.renderWindowBorder( &painter, rect(), this, client_.backgroundPalette( this, palette ) ); }
    #else
    if( NitrogenConfiguration::useCompiz() )
    {
        client_.renderWindowBackground( &painter, rect(), this, palette );

        if( client_.isActive() && client_.configuration().drawTitleOutline() && !client_.isMaximized() )
        { client_.renderWindowBorder( &painter, rect(), this, client_.backgroundPalette( this, palette ) ); }
    }
    #endif

    // draw dividing line
    painter.setRenderHints(QPainter::Antialiasing);
    QRect frame = client_.widget()->rect();
    QColor background;
    if( !( client_.configuration().overwriteColors() || client_.configuration().drawTitleOutline() ) )
    {

      background = client_.options()->color( KDecorationDefines::ColorTitleBar, client_.isActive() );

    } else { background =  palette.window().color(); }

    #if !KDE_IS_VERSION(4,3,0)
    if( client_.drawSeparator() )
    { client_.renderSeparator( &painter, rect(), this, background ); }
    #else
    if( NitrogenConfiguration::useCompiz() && client_.drawSeparator() )
    { client_.renderSeparator( &painter, rect(), this, background ); }
    #endif

    // for menu button the application icon is used
    if (type_ == ButtonMenu)
    {

      const QPixmap& pixmap( client_.icon().pixmap( client_.configuration().iconScale() ) );
      double offset = 0.5*(width()-pixmap.width() );
      painter.drawPixmap(offset, offset-1, pixmap );
      return;
    }

    // calculate color depending on timeline state
    QColor color = buttonDetailColor(palette);
    QColor glow = (type_ == ButtonClose) ?
      KColorScheme(palette.currentColorGroup()).foreground(KColorScheme::NegativeText).color():
      KColorScheme(palette.currentColorGroup()).decoration(KColorScheme::HoverColor).color();

    if( timeLineIsRunning() )
    {

      color = KColorUtils::mix( color, glow, opacity() );

    } else if( status_ == Nitrogen::Hovered ) {

      color = glow;

    }

    painter.translate( 0, -1 );

    // draw button shape
    painter.drawPixmap(0, 0, helper_.windecoButton( background, status_ == Nitrogen::Pressed, client_.configuration().buttonType(), (21.0*client_.configuration().buttonSize())/22 ) );

    // draw glow on hover
    if( timeLineIsRunning() )
    {

      painter.drawPixmap(0, 0, helper_.windecoButtonGlow( helper_.alphaColor( glow, opacity() ), (21.0*client_.configuration().buttonSize())/22));

    } else if( status_ == Nitrogen::Hovered ) {

      painter.drawPixmap(0, 0, helper_.windecoButtonGlow(glow, (21.0*client_.configuration().buttonSize())/22));

    }

    QColor decoColor( helper_.decoColor( background, color ) );

    painter.setRenderHints(QPainter::Antialiasing);
    qreal width( client_.configuration().buttonType() == NitrogenConfiguration::ButtonKde43 ? 1.4:2.2 );

    painter.setBrush(Qt::NoBrush);
    painter.setPen(QPen( decoColor, width, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
    drawIcon(&painter, palette, type_);

  }

  //___________________________________________________
  void NitrogenButton::drawIcon(QPainter *painter, QPalette &palette, ButtonType &type)
  {

    painter->save();
    painter->setWindow( 0, 0, 22, 22 );

    QPen newPen = painter->pen();
    switch(type)
    {

      case ButtonSticky:
      if(isChecked()) {
        newPen.setColor(KColorScheme(palette.currentColorGroup()).decoration(KColorScheme::HoverColor).color());
        painter->setPen(newPen);
      }
      painter->drawPoint(QPointF(10.5,10.5));
      break;

      case ButtonHelp:
      painter->translate(1.5, 1.5);
      painter->drawArc(7,5,4,4,135*16, -180*16);
      painter->drawArc(9,8,4,4,135*16,45*16);
      painter->drawPoint(9,12);
      painter->translate(-1.5, -1.5);
      break;

      case ButtonMin:
      painter->drawLine(QPointF( 7.5, 9.5), QPointF(10.5,12.5));
      painter->drawLine(QPointF(10.5,12.5), QPointF(13.5, 9.5));
      break;

      case ButtonMax:
      switch(client_.maximizeMode())
      {
        case NitrogenClient::MaximizeRestore:
        case NitrogenClient::MaximizeVertical:
        case NitrogenClient::MaximizeHorizontal:
        painter->drawLine(QPointF( 7.5,11.5), QPointF(10.5, 8.5));
        painter->drawLine(QPointF(10.5, 8.5), QPointF(13.5,11.5));
        break;

        case NitrogenClient::MaximizeFull:
        {
          painter->translate(1.5, 1.5);
          QPoint points[4] = {QPoint(9, 6), QPoint(12, 9), QPoint(9, 12), QPoint(6, 9)};
          painter->drawPolygon(points, 4);
          painter->translate(-1.5, -1.5);
          break;
        }
      }
      break;

      case ButtonClose:
      painter->drawLine(QPointF( 7.5,7.5), QPointF(13.5,13.5));
      painter->drawLine(QPointF(13.5,7.5), QPointF( 7.5,13.5));
      break;

      case ButtonAbove:
      if(isChecked()) {
        QPen newPen = painter->pen();
        newPen.setColor(KColorScheme(palette.currentColorGroup()).decoration(KColorScheme::HoverColor).color());
        painter->setPen(newPen);
      }

      painter->drawLine(QPointF( 7.5,14), QPointF(10.5,11));
      painter->drawLine(QPointF(10.5,11), QPointF(13.5,14));
      painter->drawLine(QPointF( 7.5,10), QPointF(10.5, 7));
      painter->drawLine(QPointF(10.5, 7), QPointF(13.5,10));
      break;

      case ButtonBelow:
      if(isChecked()) {
        QPen newPen = painter->pen();
        newPen.setColor(KColorScheme(palette.currentColorGroup()).decoration(KColorScheme::HoverColor).color());
        painter->setPen(newPen);
      }

      painter->drawLine(QPointF( 7.5,11), QPointF(10.5,14));
      painter->drawLine(QPointF(10.5,14), QPointF(13.5,11));
      painter->drawLine(QPointF( 7.5, 7), QPointF(10.5,10));
      painter->drawLine(QPointF(10.5,10), QPointF(13.5, 7));
      break;

      case ButtonShade:
      if (!isChecked())
      {

        // shade button
        painter->drawLine(QPointF( 7.5, 7.5), QPointF(10.5,10.5));
        painter->drawLine(QPointF(10.5,10.5), QPointF(13.5, 7.5));
        painter->drawLine(QPointF( 7.5,13.0), QPointF(13.5,13.0));

      } else {

        // unshade button
        painter->drawLine(QPointF( 7.5,10.5), QPointF(10.5, 7.5));
        painter->drawLine(QPointF(10.5, 7.5), QPointF(13.5,10.5));
        painter->drawLine(QPointF( 7.5,13.0), QPointF(13.5,13.0));

      }
      break;

      default:
      break;
    }
    painter->restore();
    return;
  }

  //_________________________________________________________________________
  void NitrogenButton::resetConfiguration( void )
  {
    timeLine_.setDuration( client_.configuration().animationsDuration() );
  }

}
