//////////////////////////////////////////////////////////////////////////////
// nitrogenclient.cpp
// -------------------
//
// Copyright (c) 2009 Hugo Pereira Da Costa <hugo.pereira@free.fr>
// Copyright (c) 2006, 2007 Casper Boemann <cbr@boemann.dk>
// Copyright (c) 2006, 2007 Riccardo Iaconelli <riccardo@kde.org>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//////////////////////////////////////////////////////////////////////////////

#include "nitrogenclient.h"
#include "nitrogenclient.moc"
#include "nitrogen.h"
#include "nitrogenbutton.h"
#include "nitrogenshadowcache.h"
#include "nitrogensizegrip.h"

#include <cassert>
#include <cmath>

#include <kdeversion.h>
#include <KLocale>
#include <KColorUtils>
#include <KDebug>
#include <KGlobal>

#include <QtGui/QLabel>
#include <QtGui/QPainter>
#include <QtGui/QApplication>


using namespace std;
namespace Nitrogen
{

  //___________________________________________
  static void oxkwincleanupBefore()
  {
    nitrogenHelper()->invalidateCaches();
    nitrogenShadowCache()->invalidateCaches();
  }

  //___________________________________________
  void renderDot(QPainter *p, const QPointF &point, qreal diameter)
  {
    p->drawEllipse(QRectF(point.x()-diameter/2, point.y()-diameter/2, diameter, diameter));
  }

  //___________________________________________
  NitrogenClient::NitrogenClient(KDecorationBridge *b, KDecorationFactory *f):
    KCommonDecorationUnstable(b, f),
    colorCacheInvalid_(true),
    size_grip_( 0 ),
    timeLine_( 200, this ),
    titleTimeLine_( 200, this ),
    helper_(*nitrogenHelper()),
    initialized_( false )
  { qAddPostRoutine(oxkwincleanupBefore); }

  //___________________________________________
  NitrogenClient::~NitrogenClient()
  {

    // delete sizegrip if any
    if( hasSizeGrip() ) deleteSizeGrip();

  }

  //___________________________________________
  QString NitrogenClient::visibleName() const
  { return i18n("Nitrogen"); }

  //___________________________________________
  void NitrogenClient::init()
  {

    KCommonDecoration::init();
    #if KDE_IS_VERSION(4,2,92)
    widget()->setAttribute(Qt::WA_NoSystemBackground );
    widget()->setAutoFillBackground( false );

    // in case of preview, one wants to make the label used
    // for the central widget transparent. This allows one to have
    // the correct background (with gradient) rendered
    // Remark: this is minor (and safe) a hack.
    // This should be moved upstream (into kwin/lib/kdecoration)
    if( isPreview() )
    {

      QList<QLabel*> children( widget()->findChildren<QLabel*>() );
      for( QList<QLabel*>::iterator iter = children.begin(); iter != children.end(); iter++ )
      { (*iter)->setAutoFillBackground( false ); }

    }

    #else
    widget()->setAttribute(Qt::WA_NoSystemBackground, !isPreview() );
    widget()->setAttribute( Qt::WA_OpaquePaintEvent, !isPreview() );
    widget()->setAutoFillBackground( isPreview() );
    #endif

    // initialize timeLine
    timeLine_.setFrameRange( 0, maxAnimationIndex );
    timeLine_.setCurveShape( QTimeLine::EaseInOutCurve );
    connect( &timeLine_, SIGNAL( frameChanged( int ) ), widget(), SLOT( update() ) );
    connect( &timeLine_, SIGNAL( finished() ), widget(), SLOT( update() ) );

    connect( &timeLine_, SIGNAL( frameChanged( int ) ), SIGNAL( dirty() ) );
    connect( &timeLine_, SIGNAL( finished() ), SIGNAL( dirty() ) );

    // initialize titleTimeLine
    titleTimeLine_.setFrameRange( 0, maxAnimationIndex );
    titleTimeLine_.setCurveShape( QTimeLine::EaseInOutCurve );
    connect( &titleTimeLine_, SIGNAL( frameChanged( int ) ), widget(), SLOT( update() ) );
    connect( &titleTimeLine_, SIGNAL( finished() ), widget(), SLOT( update() ) );
    connect( &titleTimeLine_, SIGNAL( finished() ), this, SLOT( updateOldCaption() ) );

    initialized_ = true;

    resetConfiguration();

  }

  //___________________________________________
  bool NitrogenClient::decorationBehaviour(DecorationBehaviour behaviour) const
  {
    switch (behaviour)
    {

      case DB_MenuClose:
      return true;

      case DB_WindowMask:
      return false;

      default:
      return KCommonDecoration::decorationBehaviour(behaviour);
    }
  }

  //___________________________________________
  int NitrogenClient::layoutMetric(LayoutMetric lm, bool respectWindowState, const KCommonDecorationButton *btn) const
  {

    bool maximized( isMaximized() );
    int frameBorder( configuration().frameBorder() );

    int buttonSize( configuration().hideTitleBar() ? 0 : configuration().buttonSize() );

    switch (lm)
    {
      case LM_BorderLeft:
      case LM_BorderRight:
      case LM_BorderBottom:
      {
        int border( 0 );
        if (respectWindowState && maximized)
        {

          border = 0;

        }  else if( lm == LM_BorderBottom && frameBorder >= NitrogenConfiguration::BorderNoSide ) {

          // for tiny border, the convention is to have a larger bottom area in order to
          // make resizing easier
          border = qMax(frameBorder, 4);

        }  else if( configuration().frameBorder() == NitrogenConfiguration::BorderNone && isPreview() && !compositingActive() ) {

          border = 1;

        } else if( frameBorder < NitrogenConfiguration::BorderTiny ) {

          border = 0;

        } else {

          border = frameBorder;

        }

        return border;
      }

      case LM_TitleEdgeTop:
      {
        int border = 0;
        if( frameBorder == NitrogenConfiguration::BorderNone && configuration().hideTitleBar() )
        {

            border = 0;

        } else if( !( respectWindowState && maximized ) ) {

            border = TFRAMESIZE;
            if( configuration().drawTitleOutline() ) border += HFRAMESIZE/2;

        }

        return border;

      }

      case LM_TitleEdgeBottom:
      {
        if( configuration().drawTitleOutline() ) return HFRAMESIZE/2;
        else return 0;
      }

      case LM_TitleEdgeLeft:
      case LM_TitleEdgeRight:
      {
        int border = 0;
        if( !(respectWindowState && maximized) )
        { border = 6; }

        return border;

      }

      case LM_TitleBorderLeft:
      case LM_TitleBorderRight:
      {
        int border = 5;
        if( configuration().drawTitleOutline() ) border += 2*HFRAMESIZE;
        return border;
      }

      case LM_ButtonWidth:
      case LM_ButtonHeight:
      case LM_TitleHeight:
      {
        if (respectWindowState && isToolWindow()) {

          return buttonSize;

        } else {

          return buttonSize;

        }
      }

      case LM_ButtonSpacing:
      return 1;

      case LM_ButtonMarginTop:
      return 0;


    #if KDE_IS_VERSION(4,2,92)
      // outer margin for shadow/glow
      case LM_OuterPaddingLeft:
      case LM_OuterPaddingRight:
      case LM_OuterPaddingTop:
      case LM_OuterPaddingBottom:
      if( NitrogenConfiguration::useCompiz() ) return 0;
      else return nitrogenShadowCache()->shadowSize();
    #endif

      default:
      return KCommonDecoration::layoutMetric(lm, respectWindowState, btn);
    }

  }

  //_________________________________________________________
  KCommonDecorationButton *NitrogenClient::createButton(::ButtonType type)
  {

    NitrogenButton* out;

    switch (type) {
      case MenuButton:
      out = new NitrogenButton(*this, i18n("Menu"), ButtonMenu);
      break;

      case HelpButton:
      out = new NitrogenButton(*this, i18n("Help"), ButtonHelp);
      break;

      case MinButton:
      out = new NitrogenButton(*this, i18n("Minimize"), ButtonMin);
      break;

      case MaxButton:
      out = new NitrogenButton(*this, i18n("Maximize"), ButtonMax);
      break;

      case CloseButton:
      out = new NitrogenButton(*this, i18n("Close"), ButtonClose);
      break;

      case AboveButton:
      out = new NitrogenButton(*this, i18n("Keep Above Others"), ButtonAbove);
      break;

      case BelowButton:
      out = new NitrogenButton(*this, i18n("Keep Below Others"), ButtonBelow);
      break;

      case OnAllDesktopsButton:
      out = new NitrogenButton(*this, i18n("On All Desktops"), ButtonSticky);
      break;

      case ShadeButton:
      out = new NitrogenButton(*this, i18n("Shade Button"), ButtonShade);
      break;

      default:
      out = 0;
      break;
    }

    if( out )
    {
      connect( this, SIGNAL( dirty() ), out, SLOT( update() ) );
      connect( this, SIGNAL( configurationChanged() ), out, SLOT( resetConfiguration() ) );
      out->resetConfiguration();
    }

    return out;


  }

  //_________________________________________________________
  QColor reduceContrast(const QColor &c0, const QColor &c1, double t)
  {
    double s = KColorUtils::contrastRatio(c0, c1);
    if (s < t)
      return c1;

    double l = 0.0, h = 1.0;
    double x = s, a;
    QColor r = c1;
    for (int maxiter = 16; maxiter; --maxiter)
    {

      a = 0.5 * (l + h);
      r = KColorUtils::mix(c0, c1, a);
      x = KColorUtils::contrastRatio(c0, r);

      if (fabs(x - t) < 0.01) break;
      if (x > t) h = a;
      else l = a;
    }

    return r;
  }

  //_________________________________________________________
  QRegion NitrogenClient::calcMask( void ) const
  {

    if( isMaximized() )
    { return widget()->rect(); }

    QRect frame( widget()->rect() );

    // dimensions
    int w=frame.width();
    int h=frame.height();

    // multipliers
    int left = 1;
    int right = 1;
    int top = 1;
    int bottom = 1;

    // disable bottom corners when border frame is too small and window is not shaded
    if( configuration().frameBorder() == NitrogenConfiguration::BorderNone && !isShade() ) bottom = 0;

    #if KDE_IS_VERSION(4,2,92)

    int sw = layoutMetric( LM_OuterPaddingLeft );
    int sh = layoutMetric( LM_OuterPaddingTop );

    w -= sw + layoutMetric( LM_OuterPaddingRight );
    h -= sh + layoutMetric( LM_OuterPaddingBottom );


    QRegion mask(sw + 4*left, sh + 0*top, w-4*(left+right), h-0*(top+bottom));
    mask += QRegion(sw + 0*left, sh + 4*top, w-0*(left+right), h-4*(top+bottom));
    mask += QRegion(sw + 2*left, sh + 1*top, w-2*(left+right), h-1*(top+bottom));
    mask += QRegion(sw + 1*left, sh + 2*top, w-1*(left+right), h-2*(top+bottom));

    return mask;

    #else

    int sw = 0;
    int sh = 0;

    QRegion mask(sw + 4*left, sh + 0*top, w-4*(left+right), h-0*(top+bottom));
    mask += QRegion(sw + 0*left, sh + 4*top, w-0*(left+right), h-4*(top+bottom));
    mask += QRegion(sw + 2*left, sh + 1*top, w-2*(left+right), h-1*(top+bottom));
    mask += QRegion(sw + 1*left, sh + 2*top, w-1*(left+right), h-2*(top+bottom));

    if( shadowsActive() && configuration().frameBorder() == NitrogenConfiguration::BorderNone )
    { mask += QRegion(0*left, 4*top, w-0*(left+right), h-4*(top+bottom)); }

    return mask;

    #endif

  }

  //_________________________________________________________
  QColor NitrogenClient::titlebarTextColor(const QPalette &palette)
  {
    if( timeLineIsRunning() ) return KColorUtils::mix(
      titlebarTextColor( palette, false ),
      titlebarTextColor( palette, true ),
      opacity() );
    else return titlebarTextColor( palette, isActive() );
  }

  //_________________________________________________________
  QColor NitrogenClient::titlebarTextColor(const QPalette &palette, bool active)
  {

    if( ( active && configuration().drawTitleOutline() ) || !(configuration().drawTitleOutline() || configuration().overwriteColors()) )
    {

      return options()->color(ColorFont, active );

    } else if( active ) {

      return palette.color(QPalette::Active, QPalette::WindowText);

    } else {

      if(colorCacheInvalid_)
      {

        QColor ab = palette.color(QPalette::Active, QPalette::Window);
        QColor af = palette.color(QPalette::Active, QPalette::WindowText);
        QColor nb = palette.color(QPalette::Inactive, QPalette::Window);
        QColor nf = palette.color(QPalette::Inactive, QPalette::WindowText);

        colorCacheInvalid_ = false;
        cachedTitlebarTextColor_ = reduceContrast(nb, nf, qMax(qreal(2.5), KColorUtils::contrastRatio(ab, KColorUtils::mix(ab, af, 0.4))));
      }

      return cachedTitlebarTextColor_;

    }

  }

  //_________________________________________________________
  void NitrogenClient::renderWindowBackground( QPainter* painter, const QRect& rect, const QWidget* widget, const QPalette& palette ) const
  {

    QPalette local(
      ( !configuration().overwriteColors() && !configuration().drawTitleOutline() ) ?
      backgroundPalette( widget, palette ):palette );

    if( configuration().blendColor() == NitrogenConfiguration::NoBlending )
    {

      painter->fillRect( rect, local.color( widget->window()->backgroundRole() ) );

    } else {

      #if KDE_IS_VERSION(4,2,92)
      int offset = layoutMetric( LM_OuterPaddingTop );
      #else
      int offset = 0;
      #endif

      int height = 64 + configuration().buttonSize() - 22;
      const QWidget* window( isPreview() ? NitrogenClient::widget() : widget->window() );
      helper().renderWindowBackground(painter, rect, widget, window, local, offset, height );

    }

  }


  //_________________________________________________________
  void NitrogenClient::renderWindowBorder( QPainter* painter, const QRect& clipRect, const QWidget* widget, const QPalette& palette ) const
  {

    const QWidget* window = (isPreview()) ? NitrogenClient::widget() : widget->window();

    // get coordinates relative to the client area
    // this is annoying. One could use mapTo if this was taking const QWidget* and not
    // const QWidget* as argument.
    const QWidget* w = widget;
    QPoint position( 0, 0 );
    while (  w != window && !w->isWindow() && w != w->parentWidget() ) {
      position += w->geometry().topLeft();
      w = w->parentWidget();
    }

    // save painter
    painter->save();
    if (clipRect.isValid()) {
      painter->setClipRegion(clipRect,Qt::IntersectClip);
    }

    QRect r = (isPreview()) ? NitrogenClient::widget()->rect():window->rect();

    #if KDE_IS_VERSION(4,2,92)
    if( !configuration().useCompiz() )
    {
      qreal shadowSize( nitrogenShadowCache()->shadowSize() );
      r.adjust( shadowSize, shadowSize, -shadowSize, -shadowSize ); }
    #endif

    r.adjust(0,0, 1, 1);

    // mask and painting frame
    QRegion mask;
    QRect frame;

    // top line
    {
      int shadowSize = 5;
      if( timeLineIsRunning() ) shadowSize = int(shadowSize*opacity());
      int height = HFRAMESIZE;
      QRect rect( r.topLeft()-position, QSize( r.width(), height ) );
      helper().slab( palette.color( widget->backgroundRole() ), 0, shadowSize )->render( rect.adjusted(-shadowSize-1, 0, shadowSize+1, 2 ), painter, TileSet::Bottom );
      mask += rect;
      frame |= rect;
    }

    // bottom line
    if( configuration().frameBorder() > NitrogenConfiguration::BorderNone )
    {
      int height = qMin( HFRAMESIZE, layoutMetric( LM_BorderBottom ) );
      QRect rect( r.bottomLeft()-position-QPoint(0,height), QSize( r.width(), height ) );
      mask += rect;
      frame |= rect;
    }

    // left and right
    if( configuration().frameBorder() >= NitrogenConfiguration::BorderTiny )
    {

      // left
      {
        int width = qMin( HFRAMESIZE, layoutMetric( LM_BorderLeft ) );
        QRect rect( r.topLeft()-position, QSize( width, r.height() ) );
        mask += rect;
        frame |= rect;
      }

      // right
      {
        int width = qMin( HFRAMESIZE, layoutMetric( LM_BorderRight ) );
        QRect rect( r.topRight()-position-QPoint(width,0), QSize( width, r.height() ) );
        mask += rect;
        frame |= rect;
      }

    }

    // paint
    painter->setClipRegion( mask, Qt::IntersectClip);
    renderWindowBackground(painter, frame, widget, palette );

    // restore painter
    painter->restore();

  }

  //_________________________________________________________
  void NitrogenClient::renderSeparator( QPainter* painter, const QRect& clipRect, const QWidget* widget, const QColor& color ) const
  {

    const QWidget* window = (isPreview()) ? NitrogenClient::widget() : widget->window();

    // get coordinates relative to the client area
    // this is annoying. One could use mapTo if this was taking const QWidget* and not
    // const QWidget* as argument.
    QPoint position( 0, 0 );
    {
      const QWidget* w = widget;
      while (  w != window && !w->isWindow() && w != w->parentWidget() ) {
        position += w->geometry().topLeft();
        w = w->parentWidget();
      }
    }

    if (clipRect.isValid())
    {
      painter->save();
      painter->setClipRegion(clipRect,Qt::IntersectClip);
    }

    QRect r = (isPreview()) ? NitrogenClient::widget()->rect():window->rect();
    qreal shadowSize( 0 );

    #if KDE_IS_VERSION(4,2,92)
    if( !configuration().useCompiz() ) { shadowSize = nitrogenShadowCache()->shadowSize(); }
    #endif

    r.adjust( shadowSize, shadowSize, -shadowSize, -shadowSize );
    r.adjust(0,0, 1, 1);

    // dimensions
    const int titleHeight = layoutMetric(LM_TitleHeight);
    const int titleTop = layoutMetric(LM_TitleEdgeTop) + r.top();

    // dimensions
    int x,y,w,h;
    r.getRect(&x, &y, &w, &h);

    QColor local( color );
    if( timeLineIsRunning() && configuration().separatorType() != NitrogenConfiguration::SeparatorAlways )
    { local = helper().alphaColor( color, opacity() ); }

    helper().drawSeparator( painter, QRect(x, titleTop+titleHeight-1.5, w, 2).translated( -position ), local, Qt::Horizontal);

    if (clipRect.isValid()) { painter->restore(); }

  }

  //_________________________________________________________
  void NitrogenClient::renderTitleOutline(  QPainter* painter, const QRect& rect, const QPalette& palette ) const
  {

    int shadowSize = 7;
    if( timeLineIsRunning() ) shadowSize = int(shadowSize*opacity());

    // shadow
    {
      int voffset = -shadowSize;
      if( !isMaximized() ) voffset += HFRAMESIZE;
      helper().slab( palette.color( widget()->backgroundRole() ), 0, shadowSize )->render( rect.adjusted(0, voffset, 0, 0 ), painter, TileSet::Bottom|TileSet::Left|TileSet::Right );
    }

    // center
    {
      int voffset = isMaximized() ? 0:HFRAMESIZE;
      int offset = 4;
      if( timeLineIsRunning() )  offset = int( offset*opacity() );
      renderWindowBackground(painter, rect.adjusted( offset, voffset, -offset, -offset ), widget(), palette );
    }

  }

  //_________________________________________________________
  void NitrogenClient::renderTitleText(  QPainter* painter, const QRect& rect, Qt::Alignment alignment, QColor color) const
  {
    if( titleTimeLineIsRunning() )
    {

      // if alignment is left, or right, need to get the common part of both strings
      // and draw that with full opacity and mask et out for the rest
      if( alignment & Qt::AlignLeft )
      {}

      if( !oldCaption().isEmpty() )
      {
        color.setAlphaF( 1.0 - titleOpacity() );
        painter->setPen( color );
        painter->drawText( rect, alignment, oldCaption() );
      }

      if( !caption().isEmpty() )
      {
        color.setAlphaF( titleOpacity() );
        painter->setPen( color );
        painter->drawText( rect, alignment, caption() );
      }

    } else if( !caption().isEmpty() ) {

      painter->setPen( color );
      painter->drawText( rect, alignment, caption() );

    }

  }

  //_________________________________________________________
  void NitrogenClient::activeChange( void )
  {

    // update size grip so that it gets the right color
    if( hasSizeGrip() && !( isShade() || isMaximized() ) )
    {
      sizeGrip().activeChange();
      sizeGrip().update();
    }

    // reset animation
    if( animateActiveChange() )
    {
      timeLine_.setDirection( isActive() ? QTimeLine::Forward : QTimeLine::Backward );
      if(timeLine_.state() == QTimeLine::NotRunning )
      { timeLine_.start(); }
    }

    KCommonDecorationUnstable::activeChange();

  }

  //_________________________________________________________
  void NitrogenClient::maximizeChange( void  )
  {
    if( hasSizeGrip() ) sizeGrip().setVisible( !( isShade() || isMaximized() ) );
    KCommonDecorationUnstable::maximizeChange();
  }

  //_________________________________________________________
  void NitrogenClient::shadeChange( void  )
  {
    if( hasSizeGrip() ) sizeGrip().setVisible( !( isShade() || isMaximized() ) );
    KCommonDecorationUnstable::shadeChange();
    emit dirty();
  }

  //_________________________________________________________
  void NitrogenClient::captionChange( void  )
  {
    KCommonDecorationUnstable::captionChange();

    if( !animateTitleChange() ) {
      widget()->update();
      return;
    }

    if( titleTimeLineIsRunning() ) titleTimeLine_.stop();
    titleTimeLine_.start();

  }

  //_________________________________________________________
  QPalette NitrogenClient::backgroundPalette( const QWidget* widget, QPalette palette ) const
  {

    if( timeLineIsRunning() )
    {

      QColor inactive( backgroundColor( widget, palette, false ) );
      QColor active( backgroundColor( widget, palette, true ) );
      QColor color = KColorUtils::mix( inactive, active, opacity() );
      palette.setColor( widget->window()->backgroundRole(), color );

    } else {

      palette.setColor( widget->window()->backgroundRole(), backgroundColor( widget, palette ) );

    }

    return palette;

  }

  //_________________________________________________________
  QColor NitrogenClient::backgroundColor( const QWidget* widget, QPalette palette, bool active ) const
  {

    if( ( active && configuration().drawTitleOutline() ) || !( configuration().drawTitleOutline() || configuration().overwriteColors() )  )
    {

      return options()->color( KDecorationDefines::ColorTitleBar, active );

    } else {

      return palette.color( widget->window()->backgroundRole() );

    }

  }
  //________________________________________________________________
  void NitrogenClient::drawStripes(QPainter *p, QPalette &palette, const int start, const int end, const int topMargin)
  {

    QColor color( KDecoration::options()->color(ColorTitleBar) );
    QColor light( helper_.calcLightColor(palette.color(QPalette::Window) ) );

    if( timeLineIsRunning() )
    {
      color = helper().alphaColor( color, opacity() );
      light = helper().alphaColor( light, opacity() );
    }

    QLinearGradient stripeGradient(QPoint(start,0), QPoint(end,0));
    stripeGradient.setColorAt(0.0, Qt::transparent);
    stripeGradient.setColorAt(0.05, color );
    stripeGradient.setColorAt(1.0, Qt::transparent);
    QPen pen1(stripeGradient, 0.5);

    QLinearGradient stripeGradient2(QPoint(start,0), QPoint(end,0));
    stripeGradient2.setColorAt(0.0, Qt::transparent);
    stripeGradient2.setColorAt(0.05, light);
    stripeGradient2.setColorAt(1.0, Qt::transparent);
    QPen pen2(stripeGradient2, 0.5);

    bool antialiasing = p->testRenderHint(QPainter::Antialiasing);
    p->setRenderHint(QPainter::Antialiasing, false);
    const int titleHeight = layoutMetric(LM_TitleHeight);
    double voffset = 4.0*configuration().buttonSize()/22;
    double scale = double(titleHeight-voffset)/4;
    double base = titleHeight + topMargin;

    p->save();

    for (int i = 0; i < 3; ++i)
    {

      // calculate offset for each stripe.
      int offset = int(base - voffset) - int(scale)*(i+1);
      p->setPen(pen1);
      p->drawLine(QPoint(start, offset), QPoint(end, offset));
      p->setPen(pen2);
      p->drawLine(QPoint(start, offset+1), QPoint(end, offset+1));
    }
    p->restore();
    p->setRenderHint(QPainter::Antialiasing, antialiasing);
  }

  //________________________________________________________________
  void NitrogenClient::updateWindowShape()
  {

    #if KDE_IS_VERSION(4,2,92)
    if( (isMaximized() ||  compositingActive()) && !NitrogenConfiguration::useCompiz() )
    {

      clearMask();

    } else
    #endif
    {

      setMask( calcMask() );

    }

  }

  #if !KDE_IS_VERSION(4,2,92)
  //________________________________________________________________
  QList<QRect> NitrogenClient::shadowQuads( ShadowType ) const
  {

    QSize size = widget()->size();
    int outside=21, underlap=4, cornersize=25;

    // These are underlap under the decoration so the corners look nicer 10px on the outside
    QList<QRect> quads;
    quads.append(QRect(-outside, size.height()-underlap, cornersize, cornersize));
    quads.append(QRect(underlap, size.height()-underlap, size.width()-2*underlap, cornersize));
    quads.append(QRect(size.width()-underlap, size.height()-underlap, cornersize, cornersize));
    quads.append(QRect(-outside, underlap, cornersize, size.height()-2*underlap));
    quads.append(QRect(size.width()-underlap, underlap, cornersize, size.height()-2*underlap));
    quads.append(QRect(-outside, -outside, cornersize, cornersize));
    quads.append(QRect(underlap, -outside, size.width()-2*underlap, cornersize));
    quads.append(QRect(size.width()-underlap,     -outside, cornersize, cornersize));
    return quads;

  }

  //________________________________________________________________
  double NitrogenClient::shadowOpacity( ShadowType type ) const
  {
    switch( type ) {

      case ShadowBorderedActive:
      if( isActive() ) return 1.0;
      else return 0.0;

      case ShadowBorderedInactive:
      if( isActive() )  return 0.0;
      else return 1.0;

      default:
      abort();
    }
    return 0;

  }
  #endif

  //___________________________________________
  void NitrogenClient::resetConfiguration( void )
  {

    if( !initialized_ ) return;

    configuration_ = NitrogenFactory::configuration( *this );

    // animations duration
    timeLine_.setDuration( configuration_.animationsDuration() );
    titleTimeLine_.setDuration( configuration_.animationsDuration() );

    // need to update old caption
    updateOldCaption();

    // handle size grip
    if( configuration_.drawSizeGrip() )
    {

      if( !hasSizeGrip() ) createSizeGrip();

    } else if( hasSizeGrip() ) deleteSizeGrip();

    emit configurationChanged();

  }

  //_________________________________________________________
  void NitrogenClient::paintEvent( QPaintEvent* event )
  {


    // factory
    if(!( initialized_ && NitrogenFactory::initialized() ) ) return;

    // palette
    QPalette palette = widget()->palette();
    palette.setCurrentColorGroup( (isActive() ) ? QPalette::Active : QPalette::Inactive );

    // painter
    QPainter painter(widget());
    painter.setClipRegion( event->region() );

    // define frame
    QRect frame = widget()->rect();

    // base color
    QColor color = (configuration().overwriteColors() || configuration().drawTitleOutline() || !isActive()) ?
      palette.window().color() : options()->color( ColorTitleBar, isActive());

    #if KDE_IS_VERSION(4,2,92)

    // draw shadows
    if( compositingActive() && !( NitrogenConfiguration::useCompiz() || isMaximized() ) )
    {
      TileSet *tileSet( 0 );
      if( configuration().useOxygenShadows() && timeLineIsRunning() ) tileSet = nitrogenShadowCache()->tileSet( this, timeLine_.currentFrame() );
      else tileSet = nitrogenShadowCache()->tileSet( this );

      if( !isMaximized() ) tileSet->render( frame.adjusted( 4, 4, -4, -4), &painter, TileSet::Ring);
      else if( isShade() ) tileSet->render( frame.adjusted( 0, 4, 0, -4), &painter, TileSet::Bottom);
    }

    // adjust frame
    if( !configuration().useCompiz() )
    {
      qreal shadowSize( nitrogenShadowCache()->shadowSize() );
      frame.adjust( shadowSize, shadowSize, -shadowSize, -shadowSize );
    }
    //  adjust mask
    if( (compositingActive() && !NitrogenConfiguration::useCompiz()) || isPreview() )
    {

      if( isMaximized() ) {

        painter.setClipRect( frame, Qt::IntersectClip );

      } else {

        int x, y, w, h;
        frame.getRect(&x, &y, &w, &h);

        // multipliers
        int left = 1;
        int right = 1;
        int top = 1;
        int bottom = 1;

        // disable bottom corners when border frame is too small and window is not shaded
        if( configuration().frameBorder() == NitrogenConfiguration::BorderNone && !isShade() ) bottom = 0;
        QRegion mask( x+5*left,   y+0*top, w-5*(left+right), h-0*(top+bottom));
        mask += QRegion(x+0*left, y+5*top, w-0*(left+right), h-5*(top+bottom));
        mask += QRegion(x+2*left, y+2*top, w-2*(left+right), h-2*(top+bottom));
        mask += QRegion(x+3*left, y+1*top, w-3*(left+right), h-1*(top+bottom));
        mask += QRegion(x+1*left, y+3*top, w-1*(left+right), h-3*(top+bottom));

        if( configuration().frameBorder() == NitrogenConfiguration::BorderNone )
        { mask += QRegion(x+0*left, y+4*top, w-0*(left+right), h-4*(top+bottom)); }

        painter.setClipRegion( mask, Qt::IntersectClip );

      }

    }
    #endif

    renderWindowBackground( &painter, frame, widget(), palette );
    if( ( isActive() || timeLineIsRunning() ) && configuration().drawTitleOutline() && !isMaximized() )
    {
      renderWindowBorder( &painter, frame, widget(), backgroundPalette( widget(), palette ) );
    }

    // clipping
    if( compositingActive() && !NitrogenConfiguration::useCompiz() ) painter.setClipping(false);

    // in preview mode and when frame border is 0,
    // one still draw a small rect around, unless kde is recent enough,
    // useNitrogenShadow is set to true,
    // and copositing is active
    // (that makes a lot of ifs)
    #if KDE_IS_VERSION(4,2,92)
    if(
        isPreview() &&
        ( !compositingActive() || NitrogenConfiguration::useCompiz() ) )
    #else
    if( isPreview()  && configuration().frameBorder() == NitrogenConfiguration::BorderNone )
    #endif
    {
      painter.save();
      painter.setBrush( Qt::NoBrush );
      painter.setPen( QPen( helper().calcDarkColor( widget()->palette().window().color() ), 1 ) );

      QPainterPath path;
      QPoint first( frame.topLeft()+QPoint( 0, 6 ) );
      path.moveTo( first );
      path.quadTo( frame.topLeft(), frame.topLeft()+QPoint( 6, 0 ) );
      path.lineTo( frame.topRight()-QPoint( 6, 0 ) );
      path.quadTo( frame.topRight(), frame.topRight()+QPoint( 0, 6 ) );
      path.lineTo( frame.bottomRight() );
      path.lineTo( frame.bottomLeft() );
      path.lineTo( first );
      painter.drawPath( path );
      painter.restore();

    }

    // in preview mode and if colors are not overwritten, one must redraw
    // the centerRect with the correct color
    if( isPreview() && isActive() && !( configuration().overwriteColors() || configuration().drawTitleOutline() ) )
    {
      QRect local_frame = frame.adjusted(
        layoutMetric( LM_BorderLeft )+1,
        layoutMetric( LM_TitleEdgeTop ) + layoutMetric( LM_TitleHeight ) + layoutMetric( LM_TitleEdgeBottom )+1,
        -layoutMetric( LM_BorderRight )-1,
        -layoutMetric( LM_BorderBottom )-1 );

      #if KDE_IS_VERSION(4,2,92)
      int offset = layoutMetric( LM_OuterPaddingTop );
      #else
      int offset = 0;
      #endif

      int height = 64 + configuration().buttonSize() - 22;
      const QWidget* window( isPreview() ? NitrogenClient::widget() : widget()->window() );
      helper().renderWindowBackground( &painter, local_frame, widget(), window, palette, offset, height );

    }

    // dimensions
    const int titleHeight = layoutMetric(LM_TitleHeight);
    const int titleTop = layoutMetric(LM_TitleEdgeTop) + frame.top();
    const int titleEdgeLeft = layoutMetric(LM_TitleEdgeLeft);
    const int marginLeft = layoutMetric(LM_TitleBorderLeft);
    const int marginRight = layoutMetric(LM_TitleBorderRight);

    const int titleLeft = frame.left() + titleEdgeLeft + buttonsLeftWidth() + marginLeft;
    const int titleWidth = frame.width() -
      titleEdgeLeft - layoutMetric(LM_TitleEdgeRight) -
      buttonsLeftWidth() - buttonsRightWidth() -
      marginLeft - marginRight;

    // separator
    if( drawSeparator() ) renderSeparator(&painter, frame, widget(), color );

    // draw title text
    QRect titleRect( titleLeft, titleTop-1, titleWidth, titleHeight );
    painter.setFont( options()->font(isActive(), false) );


    if( !configuration().hideTitleBar() )
    {
      if( ( isActive() || timeLineIsRunning() ) && configuration().drawTitleOutline() )
      {

        // get title bounding rect
        QRect boundingRect = painter.boundingRect( titleRect, configuration().titleAlignment() | Qt::AlignVCenter, caption() );

        // adjust
        boundingRect.setTop( frame.top() );
        boundingRect.setBottom( titleTop+titleHeight );
        boundingRect.setLeft( qMax( boundingRect.left(), titleLeft ) - 2*HFRAMESIZE );
        boundingRect.setRight( qMin( boundingRect.right(), titleLeft + titleWidth ) + 2*HFRAMESIZE );

        renderTitleOutline( &painter, boundingRect, backgroundPalette( widget(), palette ) );

      }

      // draw title text
      renderTitleText( &painter, titleRect, configuration().titleAlignment() | Qt::AlignVCenter, titlebarTextColor( backgroundPalette( widget(), palette ) ) );
    }

    painter.setRenderHint(QPainter::Antialiasing);

    // adjust if there are shadows
    #if KDE_IS_VERSION(4,2,92)
    if (compositingActive()) frame.adjust(-1,-1,1,1);
    #else
    if (shadowsActive()) frame.adjust(-1,-1, 1, 1);
    #endif

    // dimensions
    int x,y,w,h;
    frame.getRect(&x, &y, &w, &h);

    // draw stripes as indicator for active windows
    if( showStripes() )
    {

      Qt::Alignment align = configuration().titleAlignment();
      if (widget()->layoutDirection() == Qt::RightToLeft)
      {

        if (align == Qt::AlignLeft) align = Qt::AlignRight;
        else if (align == Qt::AlignRight) align = Qt::AlignLeft;

      }

      if (align & Qt::AlignLeft) {

        int left = titleLeft + QFontMetrics(options()->font(isActive(), false)).width( caption() ) + 4;
        int right = titleLeft + titleWidth;
        if( left < right )
        { drawStripes(&painter, palette, left, right, titleTop); }

      } else if (align & Qt::AlignRight) {

        int left = titleLeft;
        int right = titleLeft + titleWidth - QFontMetrics(options()->font(isActive(), false)).width( caption() ) - 4;
        if( left < right )
        { drawStripes(&painter, palette, right, left, titleTop); }

      } else if (align & Qt::AlignHCenter) {

        int textWidth = QFontMetrics(options()->font(isActive(), false)).width( caption() );
        int left = titleLeft;
        int centerLeft = titleLeft + titleWidth/2 - textWidth/2 - 4;
        int centerRight = titleLeft + titleWidth/2 + textWidth/2 + 4;
        int right = titleLeft + titleWidth;
        if( left < centerLeft ) drawStripes(&painter, palette, centerLeft, left, titleTop);
        if( centerRight < right ) drawStripes(&painter, palette, centerRight, right, titleTop);

      }

    }

    // shadow and resize handles
    #if KDE_IS_VERSION(4,2,92)
    bool drawUglyFrame = !compositingActive();
    #else
    bool drawUglyFrame = !shadowsActive();
    #endif

    if( !isMaximized() )
    {

      if( configuration().frameBorder() >= NitrogenConfiguration::BorderTiny )
      {

        helper().drawFloatFrame(
          &painter, frame, backgroundPalette( widget(), palette ).color( widget()->backgroundRole() ),
          !compositingActive(), isActive() && configuration().useOxygenShadows(),
          KDecoration::options()->color(ColorTitleBar)
          );

      } else {

        // for tiny borders, use a frame that matches the titlebar only
        QRect local( frame.topLeft(), QSize( frame.width(), layoutMetric(LM_TitleHeight) + layoutMetric(LM_TitleEdgeTop) ) );
        helper().drawFloatFrame(
          &painter, local, backgroundPalette( widget(), palette ).color( widget()->backgroundRole() ),
          false, isActive() && configuration().useOxygenShadows(),
          KDecoration::options()->color(ColorTitleBar)
          );

      }

    } else if( isShade() ) {

      // for shaded maximized windows adjust frame and draw the bottom part of it
      helper().drawFloatFrame(
        &painter, frame.adjusted( -4, 0, 4, 0 ), backgroundPalette( widget(), palette ).color( widget()->backgroundRole() ),
        drawUglyFrame, isActive(),
        KDecoration::options()->color(ColorTitleBar),
        TileSet::Bottom
        );

    }

    if( configuration().frameBorder() >= NitrogenConfiguration::BorderTiny )
    {

      if( isResizable() && !isShade() )
      {

        // Draw the 3-dots resize handles
        qreal cenY = h / 2 + x + 0.5;
        qreal posX = w + y - 2.5;
        painter.setPen(Qt::NoPen);
        painter.setBrush(QColor(0, 0, 0, 66));
        renderDot(&painter, QPointF(posX, cenY - 3), 1.8);
        renderDot(&painter, QPointF(posX, cenY), 1.8);
        renderDot(&painter, QPointF(posX, cenY + 3), 1.8);

      }

      // Draw the 3-dots resize handles
      if( isResizable() && !isShade() && !configuration().drawSizeGrip() )
      {

        painter.setPen(Qt::NoPen);
        painter.setBrush(QColor(0, 0, 0, 66));

        painter.translate(x + w-9, y + h-9);
        renderDot(&painter, QPointF(2.5, 6.5), 1.8);
        renderDot(&painter, QPointF(5.5, 5.5), 1.8);
        renderDot(&painter, QPointF(6.5, 2.5), 1.8);
      }

    }

  }

  //________________________________________________________________
  NitrogenConfiguration NitrogenClient::configuration( void ) const
  { return configuration_; }

  //_________________________________________________________________
  void NitrogenClient::createSizeGrip( void )
  {

    assert( !hasSizeGrip() );
    if( ( isResizable() && windowId() != 0 ) || isPreview() )
    {
      size_grip_ = new NitrogenSizeGrip( this );
      sizeGrip().setVisible( !( isMaximized() || isShade() ) );
    }

  }

  //_________________________________________________________________
  void NitrogenClient::deleteSizeGrip( void )
  {
    assert( hasSizeGrip() );
    size_grip_->deleteLater();
    size_grip_ = 0;
  }

}
